# UniShare.pm
#
# Author: Liraz Siri <liraz@hushmail.com>
#
# Copyright (c) 1999 Liraz Siri <liraz@hushmail.com>, Ariel, Israel
#               All rights reserved.
#
# Created: Fri Apr 16 06:33:51 IDT 1999
# Version 0.1: Fri Apr 23 09:43:35 IDT 1999

use IPC::UniShare::Base;

package IPC::UniShare::Scalar;
use Tie::VRS::Scalar;
use vars qw(@ISA $VERSION);
use Clone;

$VERSION = '0.2';
@ISA = qw(IPC::UniShare::Base Tie::VRS::Scalar);

sub init {
    my ($self, $var, @args) = @_;
    $self->new_shared(@args);
    $self->SUPER::init($var);
}

package IPC::UniShare::Array;
use Tie::VRS::Array;
use vars qw(@ISA);
@ISA = qw(IPC::UniShare::Base Tie::VRS::Array);

sub init {
    my ($self, $var, @args) = @_;
    $self->new_shared(@args);
    $self->SUPER::init($var);
}

package IPC::UniShare::Hash;
use Tie::VRS::Hash;
use vars qw(@ISA);
@ISA = qw(IPC::UniShare::Base Tie::VRS::Hash);

sub init {
    my ($self, $var, @args) = @_;
    $self->new_shared(@args);
    $self->SUPER::init($var);
}

##############################
### IPC::UniShare tie API. ###
##############################

package IPC::UniShare;

use strict;
require Exporter;
use vars qw(@EXPORT_OK @ISA);
@ISA = qw(Exporter);
@EXPORT_OK = qw(share unshare);

# share a reference, blessed (object) or otherwise.
sub share {
    my $ref = shift;

    die "'$ref' is not a reference!"
        unless ref($ref);

    # extract reference type (HASH / ARRAY / SCALAR) from object or ref.
    my $ref_type = "$ref";
    $ref_type = ($ref_type =~ /=(.+)\(/) ? $1 : ref($ref);
    
    if ($ref_type eq 'HASH') {
        tie %$ref => 'IPC::UniShare', \%$ref, @_;
    } elsif ($ref_type eq 'ARRAY') {
        tie @$ref => 'IPC::UniShare', \@$ref, @_;
    } else {
        tie $$ref => 'IPC::UniShare', $$ref, @_;
    }
    return $ref;
}

# unshare a VRS object.
sub unshare {
    my $ref = shift;

    die "'$ref' isn't even a reference!"
        unless ref($ref);

    # extract reference type (HASH / ARRAY / SCALAR) from object or ref.
    my $ref_type = "$ref";
    $ref_type = ($ref_type =~ /=(.+)\(/) ? $1 : ref($ref);

    my $tied = ($ref_type eq 'HASH' ? tied %$ref :
                $ref_type eq 'ARRAY' ? tied @$ref : tied $$ref);

    die "$ref isn't shared, can't unshare!"
        unless $tied && $tied->isa('Tie::VRS::Base');

    my $clone = Clone::dclone($tied->ref);
    untie ($ref_type eq 'HASH'  ? %$ref :
           $ref_type eq 'ARRAY' ? @$ref : $$ref);
    
    if($ref_type eq 'HASH') {
        %$ref = %$clone;
    } elsif ($ref_type eq 'ARRAY') {
        @$ref = @$clone;
    } else {
        $$ref = $$clone;
    }

    return $ref;
}

sub TIESCALAR {
    my $class = shift;
    return IPC::UniShare::Scalar->new(@_);
}

sub TIEARRAY {
    my $class = shift;
    return IPC::UniShare::Array->new(@_);
}

sub TIEHASH {
    my $class = shift;
    return IPC::UniShare::Hash->new(@_);
}

1;

__END__

=head1 NAME

IPC::UniShare - magic tie interface to shared memory.

=head1 SYNOPSIS

 # use the tie interface directly. (don't)
 use IPC::UniShare;

 tie %$object => IPC::UniShare, %$object;             # tie an object!
 tie $shared_scalar => IPC::UniShare, $shared_scalar; # tie a scalar
 tie @shared_array => IPC::UniShare, @shared_array;   # tie an array
 tie %shared_hash => IPC::UniShare, %shared_hash;     # tie a hash

 # use IPC::UniShare::share. (much better)
 use IPC::UniShare qw(share unshare);

 share($shared_object);  # don't worry, we know what kind of
                         # blessed reference your object is.

 share(\$shared_scalar); # reference to a soon-to-be-shared scalar
 share(\@shared_array);  # reference to a soon-to-be-shared array
 share(\%shared_hash);   # reference to a soon-to-be-shared hash

 if(!fork) {
           # use shared objects / hashes / arrays / scalars ...
 }
        
 unshare(\$shared_scalar);
 unshare(\@shared_array);
 # ...

=head1 DESCRIPTION

IPC::UniShare is my first stab at *real* perl support for shared
memory. By *real* I mean sharing blessed objects(!), scalars, arrays, hashes,
between processes easily and transparently.

This allows you to get away with:

=over 4

=item * poor man's threads under forks.

With UniShare, sharing objects or data between forked 'threads'
is easy and simple as pie! Unless you tell them, objects and data
won't ever know they're being shared through SysV shared memory IPC.

=item * object and data persistency.

=back
    
Star features:

=over 4

=item * Arbitrary data structures

Blessed or otherwise not limited in depth or complexity.

=item * Virtual reference system

This project's major time-sink, very tricky to implement correctly (I'll
let you know when we get there), allows shared memory to simulate
references in a consistently reallocated memory structure.

=item * Low level shm interface

Generously provided by Maurice Aubrey's IPC::ShareLite's module.

Note: Not that we're very efficient at all, the virtual reference system
'compensates' for any improvements provided by IPC::ShareLite, but just
imagine how slow we would crawl had this be implemented *entirely* in perl.

=item * Psyche friendly OO design, clean, consistent source code.

At least in my opinion.

=back

=head1 CREDITS

Many thanks to Maurice Aubrey for providing the low-level C interface
to shared memory (the B<IPC::ShareLite> module) interface without which
this module would probably be even slower. (or at least somewhat later)

And of course thanks to Ben Sugars for developing IPC::Shareable,
which provided for ... ... ... very interesting reading.

Oh, and let's not forget Raphael Manfredi who provided the always-useful
B<Storable> module. BTW, Isn't it time B<Storable> is integrated into the
standard perl distribution?

=head1 BUGS

A while ago (early pre-foobar stage) .. this module was really BAD,
inherently flawed in a crippling sort of way, now it's much better,
but several issues have yet to be resolved:

=over 4

=item * The VRS is not cooperating

The VRS does not communicate between sharing processes. While the reference
system has been designed to provide a stable simulation of references,
this is done on a per-process basis. This means it is
possible for one process to break a reference which another process
is using without notifying the process and generally cause trouble.

=item * VRS tears ties to shreads (tides tears heads from bears?)

If you give a shared variable a tie for reference, you can forget about
keeping the original tie magic intact, although VRS will slurp it's contents
into a VRS reference.

=item * Shared memory will not handle several data types:

Subreferenced objects (unblessed), Code references, GLOBS, HANDLES.
Insist on having any of these in shared objects or data structures,
and you will be branded village fool by an armed raging mob determined
to banish you to eternal exile. You've Been Warned.

=item * Infinite loop circular assignments.

share(\$share);
$share = \$share;

Don't do that!

=back

These should definitely be fixed, suggestions welcome,
patches devine.

=head1 SEE ALSO

L<IPC::ShareLite>, perlipc(1p)

=head1 AUTHOR AND COPYRIGHT

 Liraz Siri <liraz@hushmail.com>

 Copyright (c) 1999 Liraz Siri <liraz@hushmail.com>, Ariel, Israel
               All rights reserved.

 You may redistribute UniShare (including Tie::VRS) *only*
 under the GNU public license.

=cut

