# Base.pm
#
# Author: Liraz Siri <liraz@hushmail.com>
#
# Copyright (c) 1999 Liraz siri <liraz@hushmail.com>, Ariel, Israel,
#               All rights reserved.
#
# Created: Fri Apr 23 05:35:14 IDT 1999
#
# VRS base class.

package Tie::VRS::Base;
use strict;

use Devel::WeakRef::Array;
use Tie::VRS;
use Clone;

# TODO: communicate binded references between processes.
# TODO: supertie support.

################################
### Virtual Reference System ###
################################

# might be redundant, used at initialization time.
sub bond {
    my ($self, $parent, $parent_key) = @_;
    my $parent_type = $parent->type;
    $parent_type ne 'SCALAR' ?
        $parent->store($parent_key, $self->myref) :
            $parent->store($self->myref);
}

sub del_parent {
    my ($self, $parent, $key) = @_;
    my $parents = $self->parents;
    
    for (0 .. $#{$parents}) {
        my ($cache_parent, $cache_key) = ($parents->[$_]->{'object'},
                                          $parents->[$_]->{'key'});
        if($cache_parent eq $parent && $cache_key eq $key) {
            $self->myref(Clone::dclone($self->ref)) if(@$parents == 1);
            splice(@$parents, $_, 1);
        }
    }
    # TODO: if no parents left, jump to myref.
}

sub add_parent {
    my ($self, $parent, $parent_key) = @_;
    my $parents = $self->parents;
    # TODO: this will be redundant once VRS is finished,
    # no identical references should be passed.
    unless(grep($_->{'object'} eq $parent && $_->{'key'} eq $parent_key,
                @$parents)) {
        push(@$parents, {'object' => $parent, 'key' => $parent_key});
        $self->bond($parent, $parent_key);
    }
}

sub chkey_parent {
    my ($self, $parent, $old_key, $new_key) = @_;
    my $parents = $self->parents;
    foreach (@$parents) {
        $_->{'key'} = $new_key 
            if($_->{'object'} eq $parent && $_->{'key'} eq $old_key);
    }
}

sub parents {
    my $self = shift;
    return $self->{'parents'} ||= [];
}

sub myref {
    my $self = shift;
    $self->{'myref'} = shift if @_;
    unless(ref($self->{'myref'})) {
        my $var;
        my $type = $self->type;
        $self->{'myref'} = ($type eq 'HASH'  ? {} :
                            $type eq 'ARRAY' ? [] : \$var);
    }
    return $self->{'myref'};
}

sub ref {
    my $self = shift;
    my @refs = $self->refs;
    my $ref = @refs ? $refs[0] : $self->myref;
    return $ref;
}

sub refs {
    my $self = shift;
    my @refs = map($_->{'object'}->fetch($_->{'key'}),
                   @{$self->parents});
    return @refs;
}

# synchronize references.
sub sync_refs {
    my $self = shift;
    my ($ref, @refs) = $self->refs;
    foreach (@refs) {
        my $ref_type = CORE::ref($_);
        if($ref_type eq 'HASH') {
            %$_ = %$ref;
        } elsif ($ref_type eq 'ARRAY') {
            @$_ = @$ref;
        } else {
            $$_ = $$ref;
        }
    };
    foreach (@{$self->parents}) { $_->{'object'}->sync_refs() };
}

# virtual reference type. internally we would like children to know
# what kind of data type their parents are tied to.
sub type {
    my $self = shift;
    $self->{'type'} = shift if @_;
    return $self->{'type'};
}

###############
### SIGNALS ###
###############

sub changed {
    my $self = shift;
    $self->sync_refs;
    $self->signal('changed');
}

# hit our parent's start_of_iteration
sub start_of_iteration {
    my $self = shift;
    $self->signal('start_of_iteration');
}

# hit our parent's end_of_iteration.
sub end_of_iteration {
    my $self = shift;
    $self->signal('end_of_iteration');
}

#####################
### SIGNAL system ###
#####################

# The signal system is the VRS's method of communication,
# signaling reference root's of events.

sub signal {
    my ($self, $signal) = @_;
    my $parents = $self->parents;
    if(@$parents) {
        foreach (@$parents) {
            $_->{'object'}->signal($signal);
        }
    } else {
        $self->signaled($signal);
    }
}

sub signaled {
#    my $self = shift;
}

# VRS future support for superties (which are really just
# variables that were tied before we 'enchanted' them)
sub supertie {
    my $self = shift;
    $self->{'supertie'} = shift if @_;
    return $self->{'supertie'};
}

############################
### VRS magic TIE system ###
############################

### fetch the magic behind a tied reference.
###
sub tied_ref {
    my $ref = shift;
    my $type = CORE::ref($ref);
    my $tied;
    if($type eq 'HASH') {
        $tied = tied %$ref;
    } elsif ($type eq 'ARRAY') {
        $tied = tied @$ref;
    } else {
        $tied = tied $$ref;
    }
    return $tied;
}

# [$key] unreference an enchant
sub enchant_unref {
    my ($self, $key) = @_;
    my $ref = $self->enchant_ref_fetch($key);
    (tied_ref $$ref)->del_parent($self, $key) if $$ref;
    $$ref = undef;
}

# remove all enchanted references.
sub enchant_unref_all {
    my $self = shift;
    my $type = $self->type;
    my $cache = $self->enchant_ref_cache;

    my $ref;
    if($type eq 'HASH') {
        foreach my $key (keys %$cache) {
            (tied_ref $ref)->del_parent($self, $key)
                if $ref = $cache->{$key};
        }
        %$cache = ();
    } elsif ($type eq 'ARRAY') {
        foreach my $index (0 .. $#{$cache}) {
            (tied_ref $ref)->del_parent($self, $index)
                if $ref = $cache->[$index];
        }
        @$cache = ();
    } else {
        (tied_ref $$cache)->del_parent($self) if $$cache;
        $$cache = undef;
    }
}

### store a reference in our reference cache.
###
sub enchant_ref_store {
    my ($self, $var, $key) = @_;
    my $ref = $self->enchant_ref_fetch($key);
    $$ref = $var;
}

### fetch a reference from our reference cache.
###
sub enchant_ref_fetch {
    my ($self, $key) = @_;
    my $type = $self->type;
    my $cache = $self->enchant_ref_cache;
    return ($type eq 'HASH' ? \$cache->{$key} :
            $type eq 'ARRAY' ? \$cache->[$key] : \$$cache);
}

### fetch a pointer to our reference cache.
###
### REF cache's type is dependant on our data type:
### a ref_cache (hash|array|scalar) for an (hash|array|scalar) type
###
sub enchant_ref_cache {
    my $self = shift; 
    my $type = $self->type; my $var;
    $self->{'enchant_ref_cache'} ||= ($type eq 'HASH' ? {} :
                                      $type eq 'ARRAY' ? [] : \$var);
    return $self->{'enchant_ref_cache'};
}

### fetch a reference from our cache.
###
sub enchant_ref_cached {
    my ($self, $key) = @_;
    my $ref = $self->enchant_ref_fetch($key);
    return $$ref;
}

### enchant a reference, store reference in cache if ref($var)
###
sub enchant_ref {
    my ($self, $var, $key) = @_;
    if(CORE::ref($var)) {
        $self->_enchant_ref($var, $key);
        $self->enchant_ref_store($var, $key);
    }
    return $var;
}

sub _enchant_ref {
    my ($parent, $ref, $parent_key) = @_;

    my ($tied, $supertie);
    my $ref_type = CORE::ref($ref);
    if($ref_type eq 'HASH') {

        my $old_tie = tied(%$ref);
        my $old_tie_enchanted = $parent->enchanted($old_tie);
        $tied = ($old_tie_enchanted ?
                 $old_tie : tie %$ref => $parent->magic, \%$ref);
        $tied->supertie($old_tie) unless $old_tie_enchanted;

    } elsif ($ref_type eq 'ARRAY') {
        
        my $old_tie = tied(@$ref);
        my $old_tie_enchanted = $parent->enchanted($old_tie);
        $tied = ($old_tie_enchanted ?
                 $old_tie : tie @$ref => $parent->magic, \@$ref);
        $tied->supertie($old_tie) unless $old_tie_enchanted;

    } else {

        my $old_tie = tied($$ref);
        my $old_tie_enchanted = $parent->enchanted($old_tie);
        $tied = ($old_tie_enchanted ?
                 $old_tie : tie $$ref => $parent->magic, $$ref);
        $tied->supertie($old_tie) unless $old_tie_enchanted;

    }
    $tied->add_parent($parent, $parent_key);
    return $tied;
}

sub magic { 'Tie::VRS' } 
sub enchanted {
    my ($self, $object) = @_;
    my $magic = $self->magic;
    return CORE::ref($object) =~ /^${magic}::/;
}

1;
