# UniShare/Base.pm
#
# Author: Liraz Siri <liraz@hushmail.com>
#
# Copyright (c) 1999 Liraz Siri <liraz@hushmail.com>, Ariel, Israel
#               All rights reserved.
#
# Created: Fri Apr 23 05:41:23 IDT 1999
#
package IPC::UniShare::Base;
use strict;

use Clone;
use Storable qw(freeze thaw);
use IPC::ShareLite;

###############################
### SHARED MEMORY interface ###
###############################
# - shm optimization cache
# - API interface to virtual reference (store_shared, fetch_shared)

sub new_shared {
    my ($self, $arg, $mode) = @_;

    my %options = ref($arg) ? %$arg : ();
    my $key = %options ? $options{'key'} : $arg;

    # any options will override my defaults.
    my $shared = IPC::ShareLite->new(-key => $key,
                                     -mode => $mode || 0644,
                                     -create => 'yes',
                                     -destroy => $key ? 'no' : 'yes',
                                     %options)
        or die "$self->new failed: $!";
    $self->shared($shared);
    $self->shared_init();
}

sub shared {
    my $self = shift;
    $self->{'.shared'} = shift if @_;
    return $self->{'.shared'};
}

sub shared_init {
    my $self = shift;
    my $type = $self->type;
    unless($self->shared->fetch) {
        my $var;
        $self->store_shared($type eq 'HASH' ? {} :
                            $type eq 'ARRAY' ? [] : \$var);
    } else {
        my $ref_type = ref($self->fetch_shared);
        $ref_type = 'SCALAR' if $ref_type eq 'REF';
        
        my $shared_key = $self->shared->key;
        die "SHM attached to $shared_key is a $ref_type, NOT $type!"
            unless($type eq $ref_type);
    }
}

# fetch our shm reference (through the cache)
sub fetch_shared {
    my $self = shift;
    $self->shared->lock;
    unless($self->iterating ||
           ($self->shared->version == $self->cached_version)) {
        $self->cached(thaw($self->shared->fetch), $self->shared->version);
    }
    $self->shared->unlock;
    return $self->cached;
}

# store our shm reference (through the cache)
sub store_shared {
    my ($self, $shared) = @_;
    $shared = $self->cached unless $shared;
    $self->shared->lock;
    if($self->iterating) {
        $self->dirty_iteration(1);
    } else {
        $self->shared->store(freeze(Clone::dclone($shared)));
        $self->cached($shared, $self->shared->version);
    }
    $self->shared->unlock;
    return $self->cached;
}

sub cached {
    my $self = shift;
    $self->{'.cache'}->{'cached'} = shift if @_;
    $self->{'.cache'}->{'version'} = shift if @_;
    return $self->{'.cache'}->{'cached'};
}

sub cached_version {
    my $self = shift;
    $self->{'.cache'}->{'version'} = shift if @_;
    return $self->{'.cache'}->{'version'};
}

#################
### OVERLOADS ###
#################

# VRS overload:
# &Tie::VRS::Base::changed
sub changed {
    my $self = shift;
    $self->store_shared;
}

# VRS overload:
# &Tie::VRS::Base::ref
sub ref {
    my $self = shift;
    return $self->fetch_shared;
}

# overloaded IPC::Family::Base dummy.
sub signaled {
    my ($self, $signal) = @_;
    $self->$signal();
}

########################################
### Magic HASH (iteration) specifics ###
########################################

# + allow hashes to be iterated.
# + optimize access to hash.

sub start_of_iteration {
    my $self = shift;
    $self->iterating(1);
}

sub end_of_iteration {
    my $self = shift;
    $self->iterating(0);
    if($self->dirty_iteration) {
        $self->store_shared;
        $self->dirty_iteration(0);
    }
}

sub iterating {
    my $self = shift;
    $self->{'.iterating'} = shift if @_;
    return $self->{'.iterating'};
}

sub dirty_iteration {
    my $self = shift;
    $self->{'.dirty_iteration'} = shift if @_;
    return $self->{'.dirty_iteration'};
}

1;
